# Copyright 2008, Sean B. Palmer, inamidst.com
# Copyright © 2012, Elad Alfassa <elad@fedoraproject.org>
# Copyright 2012-2015, Elsie Powell, http://embolalia.com
# Copyright 2019, Florian Strzelecki <florian.strzelecki@gmail.com>
#
# Licensed under the Eiffel Forum License 2.

from __future__ import annotations

from ast import literal_eval
import inspect
import itertools
import logging
import re
import threading
import time
from types import MappingProxyType
from typing import (
    Any,
    Callable,
    Optional,
    Sequence,
    TYPE_CHECKING,
    TypeVar,
    Union,
)

from sopel import db, irc, logger, plugin, plugins, tools
from sopel.irc import modes
from sopel.lifecycle import deprecated
from sopel.plugins import (
    capabilities as plugin_capabilities,
    jobs as plugin_jobs,
    rules as plugin_rules,
)
from sopel.tools import jobs as tools_jobs
from sopel.trigger import Trigger

if TYPE_CHECKING:
    from collections.abc import Iterable, Mapping

    from sopel.plugins.handlers import AbstractPluginHandler
    from sopel.trigger import PreTrigger


__all__ = ['Sopel', 'SopelWrapper']

LOGGER = logging.getLogger(__name__)

AbstractRuleType = TypeVar('AbstractRuleType', bound=plugin_rules.AbstractRule)


class Sopel(irc.AbstractBot):
    def __init__(self, config, daemon=False):
        super().__init__(config)
        self._daemon = daemon  # Used for iPython. TODO something saner here
        self._running_triggers = []
        self._running_triggers_lock = threading.Lock()
        self._plugins: dict[str, Any] = {}
        self._rules_manager = plugin_rules.Manager()
        self._cap_requests_manager = plugin_capabilities.Manager()
        self._scheduler = plugin_jobs.Scheduler(self)

        self._url_callbacks = tools.SopelMemory()
        """Tracking of manually registered URL callbacks.

        Should be manipulated only by use of :meth:`register_url_callback` and
        :meth:`unregister_url_callback` methods, which are deprecated.

        Remove in Sopel 9, along with the above related methods.
        """

        self._times = {}
        """
        A dictionary mapping lowercased nicks to dictionaries which map
        function names to the time which they were last used by that nick.
        """

        self.modeparser = modes.ModeParser()
        """A mode parser used to parse ``MODE`` messages and modestrings."""

        self.channels = self.make_identifier_memory()
        """A map of the channels that Sopel is in.

        The keys are :class:`~sopel.tools.identifiers.Identifier`\\s of the
        channel names, and map to :class:`~sopel.tools.target.Channel` objects
        which contain the users in the channel and their permissions.
        """

        self.users = self.make_identifier_memory()
        """A map of the users that Sopel is aware of.

        The keys are :class:`~sopel.tools.identifiers.Identifier`\\s of the
        nicknames, and map to :class:`~sopel.tools.target.User` instances. In
        order for Sopel to be aware of a user, it must share at least one
        mutual channel.
        """

        self.db = db.SopelDB(config, identifier_factory=self.make_identifier)
        """The bot's database, as a :class:`sopel.db.SopelDB` instance."""

        self.memory = tools.SopelMemory()
        """
        A thread-safe dict for storage of runtime data to be shared between
        plugins. See :class:`sopel.tools.memories.SopelMemory`.
        """

        self.shutdown_methods = []
        """List of methods to call on shutdown."""

    @property
    def cap_requests(self) -> plugin_capabilities.Manager:
        """Capability Requests manager."""
        return self._cap_requests_manager

    @property
    def rules(self) -> plugin_rules.Manager:
        """Rules manager."""
        return self._rules_manager

    @property
    def scheduler(self) -> plugin_jobs.Scheduler:
        """Job Scheduler. See :func:`sopel.plugin.interval`."""
        return self._scheduler

    @property
    def command_groups(self) -> dict[str, list]:
        """A mapping of plugin names to lists of their commands.

        .. versionchanged:: 7.1
            This attribute is now generated on the fly from the registered list
            of commands and nickname commands.
        """
        # This was supposed to be deprecated, but the built-in help plugin needs it
        # TODO: create a new, better, doc interface to remove it
        plugin_commands = itertools.chain(
            self._rules_manager.get_all_commands(),
            self._rules_manager.get_all_nick_commands(),
        )
        result = {}

        for plugin_name, commands in plugin_commands:
            if plugin_name not in result:
                result[plugin_name] = list(sorted(commands.keys()))
            else:
                result[plugin_name].extend(commands.keys())
                result[plugin_name] = list(sorted(result[plugin_name]))

        return result

    @property
    def doc(self) -> dict[str, tuple]:
        """A dictionary of command names to their documentation.

        Each command is mapped to its docstring and any available examples, if
        declared in the plugin's code.

        .. versionchanged:: 3.2
            Use the first item in each callable's commands list as the key,
            instead of the function name as declared in the source code.

        .. versionchanged:: 7.1
            This attribute is now generated on the fly from the registered list
            of commands and nickname commands.
        """
        # TODO: create a new, better, doc interface to remove it
        plugin_commands = itertools.chain(
            self._rules_manager.get_all_commands(),
            self._rules_manager.get_all_nick_commands(),
        )
        commands = (
            (command, command.get_doc(), command.get_usages())
            for plugin_name, commands in plugin_commands
            for command in commands.values()
        )

        return dict(
            (name, (doc.splitlines(), [u['text'] for u in usages]))
            for command, doc, usages in commands
            for name in ((command.name,) + command.aliases)
        )

    @property
    def hostmask(self) -> Optional[str]:
        """The current hostmask for the bot :class:`~sopel.tools.target.User`.

        :return: the bot's current hostmask if the bot is connected and in
                 a least one channel; ``None`` otherwise
        """
        if not self.users or self.nick not in self.users:
            # bot must be connected and in at least one channel
            return None

        return self.users[self.nick].hostmask

    @property
    def plugins(self) -> Mapping[str, plugins.handlers.AbstractPluginHandler]:
        """A dict of the bot's currently loaded plugins.

        :return: an immutable map of plugin name to plugin object
        """
        return MappingProxyType(self._plugins)

    def has_channel_privilege(self, channel: str, privilege: int) -> bool:
        """Tell if the bot has a ``privilege`` level or above in a ``channel``.

        :param channel: a channel the bot is in
        :param privilege: privilege level to check
        :raise ValueError: when the channel is unknown

        This method checks the bot's privilege level in a channel, i.e. if it
        has this level or higher privileges::

            >>> bot.channels['#chan'].privileges[bot.nick] = plugin.OP
            >>> bot.has_channel_privilege('#chan', plugin.VOICE)
            True

        The ``channel`` argument can be either a :class:`str` or an
        :class:`~sopel.tools.identifiers.Identifier`, as long as Sopel joined
        said channel. If the channel is unknown, a :exc:`ValueError` will be
        raised.
        """
        if channel not in self.channels:
            raise ValueError('Unknown channel %s' % channel)

        return self.channels[channel].has_privilege(self.nick, privilege)

    # setup

    def setup(self) -> None:
        """Set up Sopel bot before it can run.

        The setup phase is in charge of:

        * setting up logging (configure Python's built-in :mod:`logging`)
        * setting up the bot's plugins (load, setup, and register)
        * starting the job scheduler
        """
        self.setup_logging()
        self.setup_plugins()
        self.post_setup()

    def setup_logging(self) -> None:
        """Set up logging based on config options."""
        logger.setup_logging(self.settings)
        base_format = self.settings.core.logging_format
        base_datefmt = self.settings.core.logging_datefmt

        # configure channel logging if required by configuration
        if self.settings.core.logging_channel:
            channel_level = self.settings.core.logging_channel_level
            channel_format = self.settings.core.logging_channel_format or base_format
            channel_datefmt = self.settings.core.logging_channel_datefmt or base_datefmt
            channel_params = {}
            if channel_format:
                channel_params['fmt'] = channel_format
            if channel_datefmt:
                channel_params['datefmt'] = channel_datefmt
            formatter = logger.ChannelOutputFormatter(**channel_params)
            handler = logger.IrcLoggingHandler(self, channel_level)
            handler.setFormatter(formatter)

            # set channel handler to `sopel` logger
            LOGGER = logging.getLogger('sopel')
            LOGGER.addHandler(handler)

    def setup_plugins(self) -> None:
        """Load plugins into the bot."""
        load_success = 0
        load_error = 0
        load_disabled = 0

        LOGGER.info("Loading plugins...")
        usable_plugins = plugins.get_usable_plugins(self.settings)
        for name, info in usable_plugins.items():
            plugin_handler, is_enabled = info
            if not is_enabled:
                load_disabled = load_disabled + 1
                continue

            try:
                plugin_handler.load()
            except Exception as e:
                load_error = load_error + 1
                LOGGER.exception("Error loading %s: %s", name, e)
            except SystemExit:
                load_error = load_error + 1
                LOGGER.exception(
                    "Error loading %s (plugin tried to exit)", name)
            else:
                try:
                    if plugin_handler.has_setup():
                        plugin_handler.setup(self)
                    plugin_handler.register(self)
                except Exception as e:
                    load_error = load_error + 1
                    LOGGER.exception("Error in %s setup: %s", name, e)
                else:
                    load_success = load_success + 1
                    LOGGER.info("Plugin loaded: %s", name)

        total = sum([load_success, load_error, load_disabled])
        if total and load_success:
            LOGGER.info(
                "Registered %d plugins, %d failed, %d disabled",
                (load_success - 1),
                load_error,
                load_disabled)
        else:
            LOGGER.warning("Warning: Couldn't load any plugins")

    # post setup

    def post_setup(self) -> None:
        """Perform post-setup actions.

        This method handles everything that should happen after all the plugins
        are loaded, and before the bot can connect to the IRC server.

        At the moment, this method checks for undefined configuration options,
        and starts the job scheduler.

        .. versionadded:: 7.1
        """
        settings = self.settings
        for section_name, section in settings.get_defined_sections():
            defined_options = {
                settings.parser.optionxform(opt)
                for opt, _ in inspect.getmembers(section)
                if not opt.startswith('_')
            }
            for option_name in settings.parser.options(section_name):
                if option_name not in defined_options:
                    LOGGER.warning(
                        "Config option `%s.%s` is not defined by its section "
                        "and may not be recognized by Sopel.",
                        section_name,
                        option_name,
                    )

        self._scheduler.start()

    # plugins management

    def reload_plugin(self, name: str) -> None:
        """Reload a plugin.

        :param name: name of the plugin to reload
        :raise plugins.exceptions.PluginNotRegistered: when there is no
            ``name`` plugin registered

        This function runs the plugin's shutdown routine and unregisters the
        plugin from the bot. Then this function reloads the plugin, runs its
        setup routines, and registers it again.
        """
        if not self.has_plugin(name):
            raise plugins.exceptions.PluginNotRegistered(name)

        plugin_handler = self._plugins[name]
        # tear down
        plugin_handler.shutdown(self)
        plugin_handler.unregister(self)
        LOGGER.info("Unloaded plugin %s", name)
        # reload & setup
        plugin_handler.reload()
        plugin_handler.setup(self)
        plugin_handler.register(self)
        meta = plugin_handler.get_meta_description()
        LOGGER.info("Reloaded %s plugin %s from %s",
                    meta['type'], name, meta['source'])

    def reload_plugins(self) -> None:
        """Reload all registered plugins.

        First, this function runs all plugin shutdown routines and unregisters
        all plugins. Then it reloads all plugins, runs their setup routines, and
        registers them again.
        """
        registered = list(self._plugins.items())
        # tear down all plugins
        for name, handler in registered:
            handler.shutdown(self)
            handler.unregister(self)
            LOGGER.info("Unloaded plugin %s", name)

        # reload & setup all plugins
        for name, handler in registered:
            handler.reload()
            handler.setup(self)
            handler.register(self)
            meta = handler.get_meta_description()
            LOGGER.info("Reloaded %s plugin %s from %s",
                        meta['type'], name, meta['source'])

    # TODO: deprecate both add_plugin and remove_plugin; see #2425

    def add_plugin(
        self,
        plugin: AbstractPluginHandler,
        callables: Sequence[Callable],
        jobs: Sequence[Callable],
        shutdowns: Sequence[Callable],
        urls: Sequence[Callable],
    ) -> None:
        """Add a loaded plugin to the bot's registry.

        :param plugin: loaded plugin to add
        :param callables: an iterable of callables from the ``plugin``
        :param jobs: an iterable of functions from the ``plugin`` that are
                     periodically invoked
        :param shutdowns: an iterable of functions from the ``plugin`` that
                          should be called on shutdown
        :param urls: an iterable of functions from the ``plugin`` to call when
                     matched against a URL
        """
        self._plugins[plugin.name] = plugin
        self.register_callables(callables)
        self.register_jobs(jobs)
        self.register_shutdowns(shutdowns)
        self.register_urls(urls)

    def remove_plugin(
        self,
        plugin: AbstractPluginHandler,
        callables: Sequence[Callable],
        jobs: Sequence[Callable],
        shutdowns: Sequence[Callable],
        urls: Sequence[Callable],
    ) -> None:
        """Remove a loaded plugin from the bot's registry.

        :param plugin: loaded plugin to remove
        :param callables: an iterable of callables from the ``plugin``
        :param jobs: an iterable of functions from the ``plugin`` that are
                     periodically invoked
        :param shutdowns: an iterable of functions from the ``plugin`` that
                          should be called on shutdown
        :param urls: an iterable of functions from the ``plugin`` to call when
                     matched against a URL
        """
        name = plugin.name
        if not self.has_plugin(name):
            raise plugins.exceptions.PluginNotRegistered(name)

        # remove plugin rules, jobs, shutdown functions, and url callbacks
        self._rules_manager.unregister_plugin(name)
        self._scheduler.unregister_plugin(name)
        self.unregister_shutdowns(shutdowns)

        # remove plugin from registry
        del self._plugins[name]

    def has_plugin(self, name: str) -> bool:
        """Check if the bot has registered a plugin of the specified name.

        :param str name: name of the plugin to check for
        :return: whether the bot has a plugin named ``name`` registered
        :rtype: bool
        """
        return name in self._plugins

    def get_plugin_meta(self, name: str) -> dict:
        """Get info about a registered plugin by its name.

        :param str name: name of the plugin about which to get info
        :return: the plugin's metadata
                 (see :meth:`~.plugins.handlers.AbstractPluginHandler.get_meta_description`)
        :rtype: :class:`dict`
        :raise plugins.exceptions.PluginNotRegistered: when there is no
            ``name`` plugin registered
        """
        if not self.has_plugin(name):
            raise plugins.exceptions.PluginNotRegistered(name)

        return self._plugins[name].get_meta_description()

    # callable management

    def register_callables(self, callables: Iterable) -> None:
        match_any = re.compile(r'.*')
        settings = self.settings

        for callbl in callables:
            rules = getattr(callbl, 'rule', [])
            lazy_rules = getattr(callbl, 'rule_lazy_loaders', [])
            find_rules = getattr(callbl, 'find_rules', [])
            lazy_find_rules = getattr(callbl, 'find_rules_lazy_loaders', [])
            search_rules = getattr(callbl, 'search_rules', [])
            lazy_search_rules = getattr(callbl, 'search_rules_lazy_loaders', [])
            commands = getattr(callbl, 'commands', [])
            nick_commands = getattr(callbl, 'nickname_commands', [])
            action_commands = getattr(callbl, 'action_commands', [])
            is_rule = any([
                rules,
                lazy_rules,
                find_rules,
                lazy_find_rules,
                search_rules,
                lazy_search_rules,
            ])
            is_command = any([commands, nick_commands, action_commands])

            if rules:
                rule = plugin_rules.Rule.from_callable(settings, callbl)
                self._rules_manager.register(rule)

            if lazy_rules:
                try:
                    rule = plugin_rules.Rule.from_callable_lazy(
                        settings, callbl)
                    self._rules_manager.register(rule)
                except plugins.exceptions.PluginError as err:
                    LOGGER.error('Cannot register rule: %s', err)

            if find_rules:
                rule = plugin_rules.FindRule.from_callable(settings, callbl)
                self._rules_manager.register(rule)

            if lazy_find_rules:
                try:
                    rule = plugin_rules.FindRule.from_callable_lazy(
                        settings, callbl)
                    self._rules_manager.register(rule)
                except plugins.exceptions.PluginError as err:
                    LOGGER.error('Cannot register find rule: %s', err)

            if search_rules:
                rule = plugin_rules.SearchRule.from_callable(settings, callbl)
                self._rules_manager.register(rule)

            if lazy_search_rules:
                try:
                    rule = plugin_rules.SearchRule.from_callable_lazy(
                        settings, callbl)
                    self._rules_manager.register(rule)
                except plugins.exceptions.PluginError as err:
                    LOGGER.error('Cannot register search rule: %s', err)

            if commands:
                rule = plugin_rules.Command.from_callable(settings, callbl)
                self._rules_manager.register_command(rule)

            if nick_commands:
                rule = plugin_rules.NickCommand.from_callable(
                    settings, callbl)
                self._rules_manager.register_nick_command(rule)

            if action_commands:
                rule = plugin_rules.ActionCommand.from_callable(
                    settings, callbl)
                self._rules_manager.register_action_command(rule)

            if not is_command and not is_rule:
                callbl.rule = [match_any]
                self._rules_manager.register(
                    plugin_rules.Rule.from_callable(self.settings, callbl))

    def register_jobs(self, jobs: Iterable) -> None:
        for func in jobs:
            job = tools_jobs.Job.from_callable(self.settings, func)
            self._scheduler.register(job)

    def unregister_jobs(self, jobs: Iterable) -> None:
        for job in jobs:
            self._scheduler.remove_callable_job(job)

    def register_shutdowns(self, shutdowns: Iterable) -> None:
        # Append plugin's shutdown function to the bot's list of functions to
        # call on shutdown
        self.shutdown_methods = self.shutdown_methods + list(shutdowns)

    def unregister_shutdowns(self, shutdowns: Iterable) -> None:
        self.shutdown_methods = [
            shutdown
            for shutdown in self.shutdown_methods
            if shutdown not in shutdowns
        ]

    def register_urls(self, urls: Iterable) -> None:
        for func in urls:
            url_regex = getattr(func, 'url_regex', [])
            url_lazy_loaders = getattr(func, 'url_lazy_loaders', None)

            if url_regex:
                rule = plugin_rules.URLCallback.from_callable(
                    self.settings, func)
                self._rules_manager.register_url_callback(rule)

            if url_lazy_loaders:
                try:
                    rule = plugin_rules.URLCallback.from_callable_lazy(
                        self.settings, func)
                    self._rules_manager.register_url_callback(rule)
                except plugins.exceptions.PluginError as err:
                    LOGGER.error("Cannot register URL callback: %s", err)

    def rate_limit_info(
        self,
        rule: AbstractRuleType,
        trigger: Trigger,
    ) -> tuple[bool, Optional[str]]:
        if trigger.admin or rule.is_unblockable():
            return False, None

        nick = trigger.nick
        is_channel = trigger.sender and not trigger.sender.is_nick()
        channel = trigger.sender if is_channel else None

        at_time = trigger.time
        if rule.is_user_rate_limited(nick, at_time):
            template = rule.user_rate_template
            rate_limit_type = "user"
            rate_limit = rule.user_rate_limit
            metrics = rule.get_user_metrics(nick)
        elif channel and rule.is_channel_rate_limited(channel, at_time):
            template = rule.channel_rate_template
            rate_limit_type = "channel"
            rate_limit = rule.channel_rate_limit
            metrics = rule.get_channel_metrics(channel)
        elif rule.is_global_rate_limited(at_time):
            template = rule.global_rate_template
            rate_limit_type = "global"
            rate_limit = rule.global_rate_limit
            metrics = rule.get_global_metrics()
        else:
            return False, None

        if not metrics.last_time:
            # you and I know that is_limited() will never return True if
            # last_time is None, but the type-checker doesn't
            return False, None

        next_time = metrics.last_time + rate_limit
        time_left = next_time - at_time

        message: Optional[str] = None

        if template:
            message = template.format(
                nick=trigger.nick,
                channel=channel or 'private message',
                sender=trigger.sender,
                plugin=rule.get_plugin_name(),
                label=rule.get_rule_label(),
                time_left=time_left,
                time_left_sec=time_left.total_seconds(),
                rate_limit=rate_limit,
                rate_limit_sec=rate_limit.total_seconds(),
                rate_limit_type=rate_limit_type,
            )

        return True, message

    # message dispatch

    def call_rule(
        self,
        rule: plugin_rules.AbstractRule,
        sopel: 'SopelWrapper',
        trigger: Trigger,
    ) -> None:
        nick = trigger.nick
        context = trigger.sender
        is_channel = context and not context.is_nick()

        limited, limit_msg = self.rate_limit_info(rule, trigger)
        if limit_msg:
            sopel.notice(limit_msg, destination=nick)
        if limited:
            return

        # channel config
        if is_channel and context in self.config:
            channel_config = self.config[context]
            plugin_name = rule.get_plugin_name()

            # disable listed plugins completely on provided channel
            if 'disable_plugins' in channel_config:
                disabled_plugins = channel_config.disable_plugins.split(',')

                if plugin_name == 'coretasks':
                    LOGGER.debug("disable_plugins refuses to skip a coretasks handler")
                elif '*' in disabled_plugins:
                    return
                elif plugin_name in disabled_plugins:
                    return

            # disable chosen methods from plugins
            if 'disable_commands' in channel_config:
                disabled_commands = literal_eval(channel_config.disable_commands)
                disabled_commands = disabled_commands.get(plugin_name, [])
                if rule.get_rule_label() in disabled_commands:
                    if plugin_name != 'coretasks':
                        return
                    LOGGER.debug("disable_commands refuses to skip a coretasks handler")

        try:
            rule.execute(sopel, trigger)
        except KeyboardInterrupt:
            raise
        except Exception as error:
            self.error(trigger, exception=error)

    def call(
        self,
        func: Any,
        sopel: 'SopelWrapper',
        trigger: Trigger,
    ) -> None:
        """Call a function, applying any rate limits or other restrictions.

        :param func: the function to call
        :type func: :term:`function`
        :param sopel: a SopelWrapper instance
        :type sopel: :class:`SopelWrapper`
        :param Trigger trigger: the Trigger object for the line from the server
                                that triggered this call
        """
        nick = trigger.nick
        current_time = time.time()
        if nick not in self._times:
            self._times[nick] = dict()
        if self.nick not in self._times:
            self._times[self.nick] = dict()
        if not trigger.is_privmsg and trigger.sender not in self._times:
            self._times[trigger.sender] = dict()

        if not trigger.admin and not func.unblockable:
            if func in self._times[nick]:
                usertimediff = current_time - self._times[nick][func]
                if func.user_rate > 0 and usertimediff < func.user_rate:
                    LOGGER.info(
                        "%s prevented from using %s in %s due to user limit: %d < %d",
                        trigger.nick, func.__name__, trigger.sender, usertimediff,
                        func.user_rate
                    )
                    return
            if func in self._times[self.nick]:
                globaltimediff = current_time - self._times[self.nick][func]
                if func.global_rate > 0 and globaltimediff < func.global_rate:
                    LOGGER.info(
                        "%s prevented from using %s in %s due to global limit: %d < %d",
                        trigger.nick, func.__name__, trigger.sender, globaltimediff,
                        func.global_rate
                    )
                    return

            if not trigger.is_privmsg and func in self._times[trigger.sender]:
                chantimediff = current_time - self._times[trigger.sender][func]
                if func.channel_rate > 0 and chantimediff < func.channel_rate:
                    LOGGER.info(
                        "%s prevented from using %s in %s due to channel limit: %d < %d",
                        trigger.nick, func.__name__, trigger.sender, chantimediff,
                        func.channel_rate
                    )
                    return

        # if channel has its own config section, check for excluded plugins/plugin methods,
        # but only if the source plugin is NOT coretasks, because we NEED those handlers.
        # Normal, whole-bot configuration will not let you disable coretasks either.
        if trigger.sender in self.config and func.plugin_name != 'coretasks':
            channel_config = self.config[trigger.sender]
            LOGGER.debug(
                "Evaluating configuration for %s.%s in channel %s",
                func.plugin_name, func.__name__, trigger.sender
            )

            # disable listed plugins completely on provided channel
            if 'disable_plugins' in channel_config:
                disabled_plugins = channel_config.disable_plugins.split(',')

                # if "*" is used, we are disabling all plugins on provided channel
                if '*' in disabled_plugins:
                    LOGGER.debug(
                        "All plugins disabled in %s; skipping execution of %s.%s",
                        trigger.sender, func.plugin_name, func.__name__
                    )
                    return
                if func.plugin_name in disabled_plugins:
                    LOGGER.debug(
                        "Plugin %s is disabled in %s; skipping execution of %s",
                        func.plugin_name, trigger.sender, func.__name__
                    )
                    return

            # disable chosen methods from plugins
            if 'disable_commands' in channel_config:
                disabled_commands = literal_eval(channel_config.disable_commands)

                if func.plugin_name in disabled_commands:
                    if func.__name__ in disabled_commands[func.plugin_name]:
                        LOGGER.debug(
                            "Skipping execution of %s.%s in %s: disabled_commands matched",
                            func.plugin_name, func.__name__, trigger.sender
                        )
                        return

        try:
            exit_code = func(sopel, trigger)
        except Exception as error:  # TODO: Be specific
            exit_code = None
            self.error(trigger, exception=error)

        if exit_code != plugin.NOLIMIT:
            self._times[nick][func] = current_time
            self._times[self.nick][func] = current_time
            if not trigger.is_privmsg:
                self._times[trigger.sender][func] = current_time

    def _is_pretrigger_blocked(
        self,
        pretrigger: PreTrigger,
    ) -> Union[tuple[bool, bool], tuple[None, None]]:
        if not (
            self.settings.core.nick_blocks
            or self.settings.core.host_blocks
        ):
            return (None, None)

        nick_blocked = self._nick_blocked(pretrigger.nick)
        host_blocked = self._host_blocked(pretrigger.host)
        return (nick_blocked, host_blocked)

    def dispatch(self, pretrigger: PreTrigger) -> None:
        """Dispatch a parsed message to any registered callables.

        :param pretrigger: a parsed message from the server
        :type pretrigger: :class:`~sopel.trigger.PreTrigger`

        The ``pretrigger`` (a parsed message) is used to find matching rules;
        it will retrieve them by order of priority, and execute them. It runs
        triggered rules in separate threads, unless they are marked otherwise.

        However, it won't run triggered blockable rules at all when they can't
        be executed for blocked nickname or hostname.

        .. seealso::

            The pattern matching is done by the
            :class:`Rules Manager<sopel.plugins.rules.Manager>`.

        """
        # list of commands running in separate threads for this dispatch
        running_triggers = []
        # nickname/hostname blocking
        nick_blocked, host_blocked = self._is_pretrigger_blocked(pretrigger)
        blocked = bool(nick_blocked or host_blocked)
        list_of_blocked_rules = set()
        # account info
        nick = pretrigger.nick
        user_obj = self.users.get(nick)
        account = user_obj.account if user_obj else None

        # skip processing replayed messages
        if "time" in pretrigger.tags and pretrigger.sender in self.channels:
            join_time = self.channels[pretrigger.sender].join_time
            if join_time is not None and pretrigger.time < join_time:
                return

        for rule, match in self._rules_manager.get_triggered_rules(self, pretrigger):
            trigger = Trigger(self.settings, pretrigger, match, account)

            is_unblockable = trigger.admin or rule.is_unblockable()
            if blocked and not is_unblockable:
                list_of_blocked_rules.add(str(rule))
                continue

            wrapper = SopelWrapper(
                self, trigger, output_prefix=rule.get_output_prefix())

            if rule.is_threaded():
                # run in a separate thread
                targs = (rule, wrapper, trigger)
                t = threading.Thread(target=self.call_rule, args=targs)
                plugin_name = rule.get_plugin_name()
                rule_label = rule.get_rule_label()
                t.name = '%s-%s-%s' % (t.name, plugin_name, rule_label)
                t.start()
                running_triggers.append(t)
            else:
                # direct call
                self.call_rule(rule, wrapper, trigger)

        # update currently running triggers
        self._update_running_triggers(running_triggers)

        if list_of_blocked_rules:
            if nick_blocked and host_blocked:
                block_type = 'both blocklists'
            elif nick_blocked:
                block_type = 'nick blocklist'
            else:
                block_type = 'host blocklist'
            LOGGER.debug(
                "%s prevented from using %s by %s.",
                pretrigger.nick,
                ', '.join(list_of_blocked_rules),
                block_type,
            )

    @property
    def running_triggers(self) -> list:
        """Current active threads for triggers.

        :return: the running thread(s) currently processing trigger(s)
        :rtype: :term:`iterable`

        This is for testing and debugging purposes only.
        """
        with self._running_triggers_lock:
            return [t for t in self._running_triggers if t.is_alive()]

    def _update_running_triggers(self, running_triggers: list) -> None:
        """Update list of running triggers.

        :param list running_triggers: newly started threads

        We want to keep track of running triggers, mostly for testing and
        debugging purposes. For instance, it'll help make sure, in tests, that
        a bot plugin has finished processing a trigger, by manually joining
        all running threads.

        This is kept private, as it's purely internal machinery and isn't
        meant to be manipulated by outside code.
        """
        # update bot's global running triggers
        with self._running_triggers_lock:
            running_triggers = running_triggers + self._running_triggers
            self._running_triggers = [
                t for t in running_triggers if t.is_alive()]

    # capability negotiation
    def request_capabilities(self) -> bool:
        """Request available capabilities and return if negotiation is on.

        :return: tell if the negotiation is active or not

        This takes the available capabilities and asks the request manager to
        request only these that are available.

        If none is available or if none is requested, the negotiation is not
        active and this returns ``False``. It is the responsibility of the
        caller to make sure it signals the IRC server to end the negotiation
        with a ``CAP END`` command.
        """
        available_capabilities = self._capabilities.available.keys()

        if not available_capabilities:
            LOGGER.debug('No client capability to negotiate.')
            return False

        LOGGER.info(
            "Client capability negotiation list: %s",
            ', '.join(available_capabilities),
        )

        self._cap_requests_manager.request_available(
            self, available_capabilities)

        return bool(self._cap_requests_manager.requested)

    def resume_capability_negotiation(
        self,
        cap_req: tuple[str, ...],
        plugin_name: str,
    ) -> None:
        """Resume capability negotiation and close when necessary.

        :param cap_req: a capability request
        :param plugin_name: plugin that requested the capability and wants to
                            resume capability negotiation

        This will resume a capability request through the bot's
        :attr:`capability requests manager<cap_requests>`, and if the
        negotiation wasn't completed before and is now complete, it will send
        a ``CAP END`` command.

        This method must be used by plugins that declare a capability request
        with a handler that returns
        :attr:`sopel.plugin.CapabilityNegotiation.CONTINUE` on acknowledgement
        in order for the bot to resume and eventually close negotiation.

        For example, this is useful for SASL auth which happens while
        negotiating capabilities.
        """
        was_completed, is_complete = self._cap_requests_manager.resume(
            cap_req, plugin_name,
        )
        if not was_completed and is_complete:
            LOGGER.info("End of client capability negotiation requests.")
            self.write(('CAP', 'END'))

    # event handlers

    def on_scheduler_error(
        self,
        scheduler: plugin_jobs.Scheduler,
        exc: BaseException,
    ) -> None:
        """Called when the Job Scheduler fails.

        :param scheduler: the job scheduler that errored
        :param exc: the raised exception

        .. seealso::

            :meth:`Sopel.error`
        """
        self.error(exception=exc)

    def on_job_error(
        self,
        scheduler: plugin_jobs.Scheduler,
        job: tools_jobs.Job,
        exc: BaseException,
    ) -> None:
        """Called when a job from the Job Scheduler fails.

        :param scheduler: the job scheduler responsible for the errored ``job``
        :param job: the Job that errored
        :param exc: the raised exception

        .. seealso::

            :meth:`Sopel.error`
        """
        self.error(exception=exc)

    def error(
        self,
        trigger: Optional[Trigger] = None,
        exception: Optional[BaseException] = None,
    ) -> None:
        """Called internally when a plugin causes an error.

        :param trigger: the IRC line that caused the error (if available)
        :param exception: the exception raised by the error (if available)
        """
        message = 'Unexpected error'
        if exception:
            detail = ' ({})'.format(exception) if str(exception) else ''
            message = 'Unexpected {}{}'.format(type(exception).__name__, detail)

        if trigger:
            message = '{} from {}. Message was: {}'.format(
                message, trigger.nick, trigger.group(0)
            )

        LOGGER.exception(message)

        if trigger and self.settings.core.reply_errors and trigger.sender is not None:
            self.say(message, trigger.sender)

    def _host_blocked(self, host: str) -> bool:
        """Check if a hostname is blocked.

        :param host: the hostname to check
        """
        bad_masks = self.config.core.host_blocks
        for bad_mask in bad_masks:
            bad_mask = bad_mask.strip()
            if not bad_mask:
                continue
            if (re.match(bad_mask + '$', host, re.IGNORECASE) or
                    bad_mask == host):
                return True
        return False

    def _nick_blocked(self, nick: str) -> bool:
        """Check if a nickname is blocked.

        :param nick: the nickname to check
        """
        bad_nicks = self.config.core.nick_blocks
        for bad_nick in bad_nicks:
            bad_nick = bad_nick.strip()
            if not bad_nick:
                continue
            if (re.match(bad_nick + '$', nick, re.IGNORECASE) or
                    self.make_identifier(bad_nick) == nick):
                return True
        return False

    def _shutdown(self) -> None:
        """Internal bot shutdown method."""
        LOGGER.info("Shutting down")
        # Proactively tell plugins (at least the ones that bother to check)
        self._connection_registered.clear()
        # Stop Job Scheduler
        LOGGER.info("Stopping the Job Scheduler.")
        self._scheduler.stop()

        try:
            self._scheduler.join(timeout=15)
        except RuntimeError:
            LOGGER.exception("Unable to stop the Job Scheduler.")
        else:
            LOGGER.info("Job Scheduler stopped.")

        self._scheduler.clear_jobs()

        # Shutdown plugins
        LOGGER.info(
            "Calling shutdown for %d plugins.", len(self.shutdown_methods))

        for shutdown_method in self.shutdown_methods:
            try:
                LOGGER.debug(
                    "Calling %s.%s",
                    shutdown_method.__module__,
                    shutdown_method.__name__)
                shutdown_method(self)
            except Exception as e:
                LOGGER.exception("Error calling shutdown method: %s", e)

        # Avoid calling shutdown methods if we already have.
        self.shutdown_methods = []

    # URL callbacks management

    @deprecated(
        reason='Issues with @url decorator have been fixed. Simply use that.',
        version='7.1',
        warning_in='8.0',
        removed_in='9.0',
    )
    def register_url_callback(self, pattern, callback):
        """Register a ``callback`` for URLs matching the regex ``pattern``.

        :param pattern: compiled regex pattern to register
        :type pattern: :ref:`re.Pattern <python:re-objects>`
        :param callback: callable object to handle matching URLs
        :type callback: :term:`function`

        .. versionadded:: 7.0

            This method replaces manual management of ``url_callbacks`` in
            Sopel's plugins, so instead of doing this in ``setup()``::

                if 'url_callbacks' not in bot.memory:
                    bot.memory['url_callbacks'] = tools.SopelMemory()

                regex = re.compile(r'http://example.com/path/.*')
                bot.memory['url_callbacks'][regex] = callback

            use this much more concise pattern::

                regex = re.compile(r'http://example.com/path/.*')
                bot.register_url_callback(regex, callback)

        It's recommended you completely avoid manual management of URL
        callbacks through the use of :func:`sopel.plugin.url`.

        .. deprecated:: 7.1

            Made obsolete by fixes to the behavior of
            :func:`sopel.plugin.url`. Will be removed in Sopel 9.

        .. versionchanged:: 8.0

            Stores registered callbacks in an internal property instead of
            ``bot.memory['url_callbacks']``.

        """
        if isinstance(pattern, str):
            pattern = re.compile(pattern)

        self._url_callbacks[pattern] = callback

    @deprecated(
        reason='Issues with @url decorator have been fixed. Simply use that.',
        version='7.1',
        warning_in='8.0',
        removed_in='9.0',
    )
    def unregister_url_callback(self, pattern, callback):
        """Unregister the callback for URLs matching the regex ``pattern``.

        :param pattern: compiled regex pattern to unregister callback
        :type pattern: :ref:`re.Pattern <python:re-objects>`
        :param callback: callable object to remove
        :type callback: :term:`function`

        .. versionadded:: 7.0

            This method replaces manual management of ``url_callbacks`` in
            Sopel's plugins, so instead of doing this in ``shutdown()``::

                regex = re.compile(r'http://example.com/path/.*')
                try:
                    del bot.memory['url_callbacks'][regex]
                except KeyError:
                    pass

            use this much more concise pattern::

                regex = re.compile(r'http://example.com/path/.*')
                bot.unregister_url_callback(regex, callback)

        It's recommended you completely avoid manual management of URL
        callbacks through the use of :func:`sopel.plugin.url`.

        .. deprecated:: 7.1

            Made obsolete by fixes to the behavior of
            :func:`sopel.plugin.url`. Will be removed in Sopel 9.

        .. versionchanged:: 8.0

            Deletes registered callbacks from an internal property instead of
            ``bot.memory['url_callbacks']``.

        """
        if isinstance(pattern, str):
            pattern = re.compile(pattern)

        try:
            del self._url_callbacks[pattern]
        except KeyError:
            pass

    @deprecated(
        reason=(
            'URL handling has been unified in the Rules system via the @url '
            'decorator. Use RuleManager.check_url_callbacks() if needed.'),
        version='8.0',
        warning_in='8.1',
        removed_in='9.0',
    )
    def search_url_callbacks(self, url):
        """Yield callbacks whose regex pattern matches the ``url``.

        :param str url: URL found in a trigger
        :return: yield 2-value tuples of ``(callback, match)``

        For each pattern that matches the ``url`` parameter, it yields a
        2-value tuple of ``(callable, match)`` for that pattern.

        The ``callable`` is the one registered with
        :meth:`register_url_callback`, and the ``match`` is the result of
        the regex pattern's ``search`` method.

        .. versionadded:: 7.0

        .. versionchanged:: 8.0

            Searches for registered callbacks in an internal property instead
            of ``bot.memory['url_callbacks']``.

        .. deprecated:: 8.0

            Made obsolete by fixes to the behavior of
            :func:`sopel.plugin.url`. Will be removed in Sopel 9.

        .. seealso::

            The Python documentation for the `re.search`__ function and
            the `match object`__.

        .. __: https://docs.python.org/3.11/library/re.html#re.search
        .. __: https://docs.python.org/3.11/library/re.html#match-objects

        """
        for regex, function in self._url_callbacks.items():
            match = regex.search(url)
            if match:
                yield function, match


class SopelWrapper:
    """Wrapper around a Sopel instance and a Trigger.

    :param sopel: Sopel instance
    :type sopel: :class:`~sopel.bot.Sopel`
    :param trigger: IRC Trigger line
    :type trigger: :class:`~sopel.trigger.Trigger`
    :param str output_prefix: prefix for messages sent through this wrapper
                              (e.g. plugin tag)

    This wrapper will be used to call Sopel's triggered commands and rules as
    their ``bot`` argument. It acts as a proxy, providing the ``trigger``'s
    ``sender`` (source channel or private message) as the default
    ``destination`` argument for overridden methods.

    .. deprecated:: 8.0

        ``SopelWrapper`` will be replaced with a ``contextvars`` based
        alternative. For more information, see :issue:`2460`.

    """
    def __init__(self, sopel, trigger, output_prefix=''):
        if not output_prefix:
            # Just in case someone passes in False, None, etc.
            output_prefix = ''
        # The custom __setattr__ for this class sets the attribute on the
        # original bot object. We don't want that for these, so we set them
        # with the normal __setattr__.
        object.__setattr__(self, '_bot', sopel)
        object.__setattr__(self, '_trigger', trigger)
        object.__setattr__(self, '_out_pfx', output_prefix)

    def __dir__(self):
        classattrs = [attr for attr in self.__class__.__dict__
                      if not attr.startswith('__')]
        return list(self.__dict__) + classattrs + dir(self._bot)

    def __getattr__(self, attr):
        return getattr(self._bot, attr)

    def __setattr__(self, attr, value):
        return setattr(self._bot, attr, value)

    @property
    def default_destination(self) -> Optional[str]:
        """Default say/reply destination for the associated Trigger.

        :return: the channel (with status prefix) or nick to send messages to

        This property returns the :class:`str` version of the destination that
        will be used by default by these methods:

        * :meth:`say`
        * :meth:`reply`
        * :meth:`action`
        * :meth:`notice`

        For a channel, it also ensures that the status-specific prefix is added
        to the result, so the bot replies with the same status.
        """
        if not self._trigger.sender:
            return None

        # ensure str and not Identifier
        destination = str(self._trigger.sender)

        # prepend status prefix if it exists
        if self._trigger.status_prefix:
            destination = self._trigger.status_prefix + destination

        return destination

    def say(self, message, destination=None, max_messages=1, truncation='', trailing=''):
        """Override ``Sopel.say`` to use trigger source by default.

        :param str message: message to say
        :param str destination: channel or nickname; defaults to
            :attr:`trigger.sender <sopel.trigger.Trigger.sender>`
        :param int max_messages: split ``message`` into at most this many
                                 messages if it is too long to fit into one
                                 line (optional)
        :param str truncation: string to indicate that the ``message`` was
                               truncated (optional)
        :param str trailing: string that should always appear at the end of
                             ``message`` (optional)

        The ``destination`` will default to the channel in which the
        trigger happened (or nickname, if received in a private message).

        .. seealso::

            For more details about the optional arguments to this wrapper
            method, consult the documentation for :meth:`sopel.bot.Sopel.say`.

        """
        if destination is None:
            destination = self.default_destination

        self._bot.say(
            self._out_pfx + message,
            destination,
            max_messages,
            truncation,
            trailing,
        )

    def action(self, message, destination=None):
        """Override ``Sopel.action`` to use trigger source by default.

        :param str message: action message
        :param str destination: channel or nickname; defaults to
            :attr:`trigger.sender <sopel.trigger.Trigger.sender>`

        The ``destination`` will default to the channel in which the
        trigger happened (or nickname, if received in a private message).

        .. seealso::

            :meth:`sopel.bot.Sopel.action`
        """
        if destination is None:
            destination = self.default_destination

        self._bot.action(message, destination)

    def notice(self, message, destination=None):
        """Override ``Sopel.notice`` to use trigger source by default.

        :param str message: notice message
        :param str destination: channel or nickname; defaults to
            :attr:`trigger.sender <sopel.trigger.Trigger.sender>`

        The ``destination`` will default to the channel in which the
        trigger happened (or nickname, if received in a private message).

        .. seealso::

            :meth:`sopel.bot.Sopel.notice`
        """
        if destination is None:
            destination = self.default_destination

        self._bot.notice(self._out_pfx + message, destination)

    def reply(self, message, destination=None, reply_to=None, notice=False):
        """Override ``Sopel.reply`` to ``reply_to`` sender by default.

        :param str message: reply message
        :param str destination: channel or nickname; defaults to
            :attr:`trigger.sender <sopel.trigger.Trigger.sender>`
        :param str reply_to: person to reply to; defaults to
            :attr:`trigger.nick <sopel.trigger.Trigger.nick>`
        :param bool notice: reply as an IRC notice or with a simple message

        The ``destination`` will default to the channel in which the
        trigger happened (or nickname, if received in a private message).

        ``reply_to`` will default to the nickname who sent the trigger.

        .. seealso::

            :meth:`sopel.bot.Sopel.reply`
        """
        if destination is None:
            destination = self.default_destination

        if reply_to is None:
            reply_to = self._trigger.nick

        self._bot.reply(message, destination, reply_to, notice)

    def kick(self, nick, channel=None, message=None):
        """Override ``Sopel.kick`` to kick in a channel

        :param str nick: nick to kick out of the ``channel``
        :param str channel: optional channel to kick ``nick`` from
        :param str message: optional message for the kick

        The ``channel`` will default to the channel in which the call was
        triggered. If triggered from a private message, ``channel`` is
        required.

        .. seealso::

            :meth:`sopel.bot.Sopel.kick`
        """
        if channel is None:
            if self._trigger.is_privmsg:
                raise RuntimeError('Error: KICK requires a channel.')
            else:
                channel = self._trigger.sender

        if nick is None:
            raise RuntimeError('Error: KICK requires a nick.')

        self._bot.kick(nick, channel, message)



"""
calc.py - Sopel Calculator Plugin
Copyright 2008, Sean B. Palmer, inamidst.com
Licensed under the Eiffel Forum License 2.

https://sopel.chat
"""
from __future__ import annotations

from sopel import plugin
from sopel.tools.calculation import eval_equation


@plugin.command('c', 'calc')
@plugin.example('.c', 'Nothing to calculate.')
@plugin.example('.c foo * bar', "Can't process expression: Node type 'Name' is not supported.")
@plugin.example('.c 10 / 0', 'Division by zero is not supported in this universe.')
@plugin.example('.c 10\\2', 'Invalid syntax')
@plugin.example('.c (10**1000000)**2',
                "Error running calculation: "
                "ValueError('Pow expression too complex to calculate.')")
@plugin.example('.c (10**100000)**2 * (10**100000)**2',
                "Error running calculation: "
                "ValueError('Value is too large to be handled in limited time and memory.')")
@plugin.example('.c 5 + 3', '8')
@plugin.example('.c 0.9*10', '9')
@plugin.example('.c 10*0.9', '9')
@plugin.example('.c 0.5**2', '0.25')
@plugin.example('.c 3**0', '1')
@plugin.example('.c 0 * 5', '0')
@plugin.example('.c 5**5', '3125')
@plugin.example('.c 2*(1+2)*3', '18', user_help=True)
@plugin.example('.c 2**10', '1024', user_help=True)
@plugin.example('.c 5 // 2', '2', user_help=True)
@plugin.example('.c 5 / 2', '2.5', user_help=True)
@plugin.output_prefix('[calc] ')
def c(bot, trigger):
    """Evaluate some calculation."""
    if not trigger.group(2):
        bot.reply('Nothing to calculate.')
        return
    # Account for the silly non-Anglophones and their silly radix point.
    eqn = trigger.group(2).replace(',', '.')
    try:
        result = "{:.10g}".format(eval_equation(eqn))
    except eval_equation.Error as err:
        bot.reply("Can't process expression: {}".format(str(err)))
        return
    except ZeroDivisionError:
        bot.reply('Division by zero is not supported in this universe.')
        return
    except SyntaxError:
        bot.reply('Invalid syntax')
        return
    except ValueError as err:
        bot.reply("Error running calculation: %r" % err)
        return

    bot.say(result)



"""Sopel Plugins Command Line Interface (CLI): ``sopel-plugins``"""
from __future__ import annotations

import argparse
import inspect
import operator

from sopel import config, plugins
from . import utils


ERR_CODE = 1
"""Error code: program exited with an error"""


def build_parser():
    """Configure an argument parser for ``sopel-plugins``.

    :return: the argument parser
    :rtype: :class:`argparse.ArgumentParser`
    """
    parser = argparse.ArgumentParser(
        description='Sopel plugins tool')

    # Subparser: sopel-plugins <sub-parser> <sub-options>
    subparsers = parser.add_subparsers(
        help='Action to perform',
        dest='action')

    # sopel-plugins show <name>
    show_parser = subparsers.add_parser(
        'show',
        formatter_class=argparse.RawTextHelpFormatter,
        help="Show plugin details",
        description="Show detailed information about a plugin.")
    utils.add_common_arguments(show_parser)
    show_parser.add_argument('name', help='Plugin name')

    # sopel-plugins configure <name>
    config_parser = subparsers.add_parser(
        'configure',
        formatter_class=argparse.RawTextHelpFormatter,
        help="Configure plugin with a config wizard",
        description=inspect.cleandoc("""
            Run a config wizard to configure a plugin.

            This can be used whether the plugin is enabled or not.
        """))
    utils.add_common_arguments(config_parser)
    config_parser.add_argument('name', help='Plugin name')

    # sopel-plugins list
    list_parser = subparsers.add_parser(
        'list',
        formatter_class=argparse.RawTextHelpFormatter,
        help="List available Sopel plugins",
        description=inspect.cleandoc("""
            List available Sopel plugins from all possible sources.

            Plugin sources are: built-in, from ``sopel_modules.*``,
            from ``sopel.plugins`` entry points, or Sopel's plugin directories.

            Enabled plugins are displayed in green; disabled, in red.
        """))
    utils.add_common_arguments(list_parser)
    list_parser.add_argument(
        '-C', '--no-color',
        help='Disable colors',
        dest='no_color',
        action='store_true',
        default=False)
    list_enable = list_parser.add_mutually_exclusive_group(required=False)
    list_enable.add_argument(
        '-e', '--enabled-only',
        help='Display only enabled plugins',
        dest='enabled_only',
        action='store_true',
        default=False)
    list_enable.add_argument(
        '-d', '--disabled-only',
        help='Display only disabled plugins',
        dest='disabled_only',
        action='store_true',
        default=False)
    list_parser.add_argument(
        '-n', '--name-only',
        help='Display only plugin names',
        dest='name_only',
        action='store_true',
        default=False)

    # sopel-plugins disable
    disable_parser = subparsers.add_parser(
        'disable',
        formatter_class=argparse.RawTextHelpFormatter,
        help="Disable a Sopel plugins",
        description=inspect.cleandoc("""
            Disable a Sopel plugin by its name, no matter where it comes from.

            It is not possible to disable the ``coretasks`` plugin.
        """))
    utils.add_common_arguments(disable_parser)
    disable_parser.add_argument(
        'names', metavar='name', nargs='+',
        help=inspect.cleandoc("""
            Name of the plugin to disable.
            Can be used multiple times to disable multiple plugins at once.
            In case of error, configuration is not modified.
        """))
    disable_parser.add_argument(
        '-f', '--force', action='store_true', default=False,
        help=inspect.cleandoc("""
            Force exclusion of the plugin.
            When ``core.enable`` is defined, a plugin may be disabled without
            being excluded. In this case, use this option to force
            its exclusion.
        """))
    disable_parser.add_argument(
        '-r', '--remove', action='store_true', default=False,
        help="Remove from ``core.enable`` list if applicable.")

    # sopel-plugins enable
    enable_parser = subparsers.add_parser(
        'enable',
        formatter_class=argparse.RawTextHelpFormatter,
        help="Enable a Sopel plugin",
        description=inspect.cleandoc("""
            Enable a Sopel plugin by its name, no matter where it comes from.

            The ``coretasks`` plugin is always enabled.

            By default, a plugin that is not excluded is enabled, unless at
            least one plugin is defined in the ``core.enable`` list.
            In that case, Sopel uses an "allow-only" policy for plugins, and
            all desired plugins must be added to this list.
        """))
    utils.add_common_arguments(enable_parser)
    enable_parser.add_argument(
        'names', metavar='name', nargs='+',
        help=inspect.cleandoc("""
            Name of the plugin to enable.
            Can be used multiple times to enable multiple plugins at once.
            In case of error, configuration is not modified.
        """))
    enable_parser.add_argument(
        '-a', '--allow-only',
        dest='allow_only',
        action='store_true',
        default=False,
        help=inspect.cleandoc("""
            Enforce allow-only policy.
            It makes sure the plugin is added to the ``core.enable`` list.
        """))

    return parser


def handle_list(options):
    """List Sopel plugins.

    :param options: parsed arguments
    :type options: :class:`argparse.Namespace`
    :return: 0 if everything went fine
    """
    settings = utils.load_settings(options)
    no_color = options.no_color
    name_only = options.name_only
    enabled_only = options.enabled_only
    disabled_only = options.disabled_only

    # get usable plugins
    items = (
        (name, info[0], info[1])
        for name, info in plugins.get_usable_plugins(settings).items()
    )
    items = (
        (name, plugin, is_enabled)
        for name, plugin, is_enabled in items
    )
    # filter on enabled/disabled if required
    if enabled_only:
        items = (
            (name, plugin, is_enabled)
            for name, plugin, is_enabled in items
            if is_enabled
        )
    elif disabled_only:
        items = (
            (name, plugin, is_enabled)
            for name, plugin, is_enabled in items
            if not is_enabled
        )
    # sort plugins
    sorted_plugins = sorted(items, key=operator.itemgetter(0))

    for name, plugin, is_enabled in sorted_plugins:
        description = {
            'name': name,
            'status': 'enabled' if is_enabled else 'disabled',
        }

        # optional meta description from the plugin itself
        try:
            plugin.load()
            description.update(plugin.get_meta_description())

            # colorize name for display purpose
            if not no_color:
                if is_enabled:
                    description['name'] = utils.green(name)
                else:
                    description['name'] = utils.red(name)
        except Exception as error:
            description.update(plugin.get_meta_description())

            meta_label = description.get('label') or '(error)'
            error_label = ('%s' % error) or 'unknown loading exception'
            label = '%s (error: %s)' % (meta_label, error_label)
            error_status = 'error'

            description.update({
                'label': label,
                'status': error_status,
            })
            if not no_color:
                if is_enabled:
                    # yellow instead of green
                    description['name'] = utils.yellow(name)
                else:
                    # keep it red for disabled plugins
                    description['name'] = utils.red(name)
                description['status'] = utils.red(error_status)

        template = '{name}/{type} {label} ({source}) [{status}]'
        if name_only:
            template = '{name}'

        print(template.format(**description))

    return 0  # successful operation


def handle_show(options):
    """Show plugin details.

    :param options: parsed arguments
    :type options: :class:`argparse.Namespace`
    :return: 0 if everything went fine;
             1 if the plugin doesn't exist or can't be loaded
    """
    plugin_name = options.name
    settings = utils.load_settings(options)
    usable_plugins = plugins.get_usable_plugins(settings)

    # plugin does not exist
    if plugin_name not in usable_plugins:
        utils.stderr('No plugin named %s' % plugin_name)
        return ERR_CODE

    plugin, is_enabled = usable_plugins[plugin_name]
    description = {
        'name': plugin_name,
        'status': 'enabled' if is_enabled else 'disabled',
    }

    # optional meta description from the plugin itself
    loaded = False
    try:
        plugin.load()
        description.update(plugin.get_meta_description())
        loaded = True
    except Exception as error:
        description.update(plugin.get_meta_description())

        meta_label = description.get('label') or '(error)'
        error_label = ('%s' % error) or 'unknown loading exception'
        label = '%s (error: %s)' % (meta_label, error_label)
        error_status = 'error'

        description.update({
            'label': label,
            'status': error_status,
        })

    print('Plugin:', description['name'])
    print('Version:', description['version'] or 'unknown')
    print('Status:', description['status'])
    print('Type:', description['type'])
    print('Source:', description['source'])
    print('Label:', description['label'])

    if not loaded:
        print('Loading failed')
        return ERR_CODE

    print('Loaded successfully')
    print('Setup:', 'yes' if plugin.has_setup() else 'no')
    print('Shutdown:', 'yes' if plugin.has_shutdown() else 'no')
    print('Configure:', 'yes' if plugin.has_configure() else 'no')
    return 0  # successful operation


def handle_configure(options):
    """Configure a Sopel plugin with a config wizard.

    :param options: parsed arguments
    :type options: :class:`argparse.Namespace`
    :return: 0 if everything went fine;
             1 if the plugin doesn't exist or can't be loaded
    """
    plugin_name = options.name
    settings = utils.load_settings(options)
    usable_plugins = plugins.get_usable_plugins(settings)

    # plugin does not exist
    if plugin_name not in usable_plugins:
        utils.stderr('No plugin named %s' % plugin_name)
        return ERR_CODE

    plugin, is_enabled = usable_plugins[plugin_name]
    try:
        plugin.load()
    except Exception as error:
        utils.stderr('Cannot load plugin %s: %s' % (plugin_name, error))
        return ERR_CODE

    if not plugin.has_configure():
        utils.stderr('Nothing to configure for plugin %s' % plugin_name)
        return 0  # nothing to configure is not exactly an error case

    print('Configure %s' % plugin.get_label())
    try:
        plugin.configure(settings)
    except KeyboardInterrupt:
        utils.stderr(
            '\nOperation cancelled; the config file has not been modified.')
        return ERR_CODE  # cancelled operation

    settings.save()

    if not is_enabled:
        utils.stderr(
            "Plugin {0} has been configured but is not enabled. "
            "Use 'sopel-plugins enable {0}' to enable it".format(plugin_name)
        )

    return 0  # successful operation


def _handle_disable_plugin(settings, plugin_name, force):
    excluded = settings.core.exclude
    # nothing left to do if already excluded
    if plugin_name in excluded:
        utils.stderr('Plugin %s already disabled.' % plugin_name)
        return False

    # recalculate state: at the moment, the plugin is not in the excluded list
    # however, with ensure_remove, the enable list may be empty, so we have
    # to compute the plugin's state here, and not use what comes from
    # plugins.get_usable_plugins
    is_enabled = (
        not settings.core.enable or
        plugin_name in settings.core.enable
    )

    # if not enabled at this point, exclude if options.force is used
    if not is_enabled and not force:
        utils.stderr(
            'Plugin %s is disabled but not excluded; '
            'use -f/--force to force its exclusion.'
            % plugin_name)
        return False

    settings.core.exclude = excluded + [plugin_name]
    return True


def display_unknown_plugins(unknown_plugins):
    """Print an error message when one or more plugins are unknown.

    :param list unknown_plugins: list of unknown plugins
    """
    # at least one of the plugins does not exist
    utils.stderr(utils.get_many_text(
        unknown_plugins,
        one='No plugin named {item}.',
        two='No plugin named {first} or {second}.',
        many='No plugin named {left}, or {last}.'
    ))


def handle_disable(options):
    """Disable Sopel plugins.

    :param options: parsed arguments
    :type options: :class:`argparse.Namespace`
    :return: 0 if everything went fine;
             1 if the plugin doesn't exist,
             or if attempting to disable coretasks (required)
    """
    plugin_names = options.names
    force = options.force
    ensure_remove = options.remove
    settings = utils.load_settings(options)
    usable_plugins = plugins.get_usable_plugins(settings)

    # coretasks is sacred
    if 'coretasks' in plugin_names:
        utils.stderr('Plugin coretasks cannot be disabled.')
        return ERR_CODE  # do nothing and return an error code

    unknown_plugins = [
        name
        for name in plugin_names
        if name not in usable_plugins
    ]
    if unknown_plugins:
        display_unknown_plugins(unknown_plugins)
        return ERR_CODE  # do nothing and return an error code

    # remove from enabled if asked
    if ensure_remove:
        settings.core.enable = [
            name
            for name in settings.core.enable
            if name not in plugin_names
        ]
        settings.save()

    # disable plugin (when needed)
    actually_disabled = tuple(
        name
        for name in plugin_names
        if _handle_disable_plugin(settings, name, force)
    )

    # save if required
    if actually_disabled:
        settings.save()
    else:
        return 0  # nothing to disable or save, but not an error case

    # display plugins actually disabled by the command
    print(utils.get_many_text(
        actually_disabled,
        one='Plugin {item} disabled.',
        two='Plugins {first} and {second} disabled.',
        many='Plugins {left}, and {last} disabled.'
    ))

    return 0


def _handle_enable_plugin(settings, usable_plugins, plugin_name, allow_only):
    enabled = settings.core.enable
    excluded = settings.core.exclude

    # coretasks is sacred
    if plugin_name == 'coretasks':
        utils.stderr('Plugin coretasks is always enabled.')
        return False

    # is it already enabled, but should we enforce anything?
    is_enabled = usable_plugins[plugin_name][1]
    if is_enabled and not allow_only:
        # already enabled, and no allow-only option: all good
        if plugin_name in enabled:
            utils.stderr('Plugin %s is already enabled.' % plugin_name)
        else:
            # suggest to use --allow-only option
            utils.stderr(
                'Plugin %s is enabled; '
                'use option -a/--allow-only to enforce allow only policy.'
                % plugin_name)

        return False

    # not enabled, or option allow_only to enforce
    if plugin_name in excluded:
        # remove from excluded
        settings.core.exclude = [
            name
            for name in settings.core.exclude
            if plugin_name != name
        ]
    elif plugin_name in enabled:
        # not excluded, and already in enabled list: all good
        utils.stderr('Plugin %s is already enabled' % plugin_name)
        return False

    if plugin_name not in enabled and (enabled or allow_only):
        # not excluded, but not enabled either: allow-only mode required
        # either because of the current configuration, or by request
        settings.core.enable = enabled + [plugin_name]

    return True


def handle_enable(options):
    """Enable a Sopel plugin.

    :param options: parsed arguments
    :type options: :class:`argparse.Namespace`
    :return: 0 if everything went fine;
             1 if the plugin doesn't exist
    """
    plugin_names = options.names
    allow_only = options.allow_only
    settings = utils.load_settings(options)
    usable_plugins = plugins.get_usable_plugins(settings)

    # plugin does not exist
    unknown_plugins = [
        name
        for name in plugin_names
        if name not in usable_plugins
    ]
    if unknown_plugins:
        display_unknown_plugins(unknown_plugins)
        return ERR_CODE  # do nothing and return an error code

    actually_enabled = tuple(
        name
        for name in plugin_names
        if _handle_enable_plugin(settings, usable_plugins, name, allow_only)
    )

    # save if required
    if actually_enabled:
        settings.save()
    else:
        return 0  # nothing to disable or save, but not an error case

    # display plugins actually disabled by the command
    print(utils.get_many_text(
        actually_enabled,
        one='Plugin {item} enabled.',
        two='Plugins {first} and {second} enabled.',
        many='Plugins {left}, and {last} enabled.'
    ))
    return 0


def main():
    """Console entry point for ``sopel-plugins``."""
    parser = build_parser()
    options = parser.parse_args()
    action = options.action

    if not action:
        parser.print_help()
        return ERR_CODE

    try:
        if action == 'list':
            return handle_list(options)
        elif action == 'show':
            return handle_show(options)
        elif action == 'configure':
            return handle_configure(options)
        elif action == 'disable':
            return handle_disable(options)
        elif action == 'enable':
            return handle_enable(options)
    except KeyboardInterrupt:
        utils.stderr('Bye!')
        return ERR_CODE
    except config.ConfigurationNotFound as err:
        utils.stderr(err)
        utils.stderr('Use `sopel-config init` to create a new config file.')
        return ERR_CODE



""":mod:`sopel.irc.abstract_backends` defines the IRC backend interface.

.. warning::

    This is all internal code, not intended for direct use by plugins. It is
    subject to change between versions, even patch releases, without any
    advance warning.

    Please use the public APIs on :class:`bot <sopel.bot.Sopel>`.

"""
# Copyright 2019, Florian Strzelecki <florian.strzelecki@gmail.com>
#
# Licensed under the Eiffel Forum License 2.
from __future__ import annotations

import abc
import logging
from typing import Optional, TYPE_CHECKING

from .utils import safe


if TYPE_CHECKING:
    from sopel.irc import AbstractBot
    from sopel.trigger import PreTrigger


class AbstractIRCBackend(abc.ABC):
    """Abstract class defining the interface and basic logic of an IRC backend.

    :param bot: a Sopel instance
    :type bot: :class:`sopel.bot.Sopel`

    Some methods of this class **MUST** be overridden by a subclass, or the
    backend implementation will not function correctly.
    """
    def __init__(self, bot: AbstractBot):
        self.bot: AbstractBot = bot

    @abc.abstractmethod
    def is_connected(self) -> bool:
        """Tell if the backend is connected or not."""

    def log_exception(self) -> None:
        """Log an exception to ``sopel.exceptions``.

        The IRC backend must use this method to log any exception that isn't
        caught by the bot itself (i.e. while handling messages), such as
        connection errors, SSL errors, etc.
        """
        err_log = logging.getLogger('sopel.exceptions')
        err_log.exception('Exception in core')
        err_log.error('----------------------------------------')

    @abc.abstractmethod
    def on_irc_error(self, pretrigger: PreTrigger) -> None:
        """Action to perform when the server sends an error event.

        :param pretrigger: PreTrigger object with the error event

        On IRC error, if ``bot.hasquit`` is set, the backend should close the
        connection so the bot can quit or reconnect as required.
        """

    @abc.abstractmethod
    def irc_send(self, data: bytes) -> None:
        """Send an IRC line as raw ``data``.

        :param bytes data: raw line to send

        This method must be thread-safe.
        """

    @abc.abstractmethod
    def run_forever(self) -> None:
        """Run the backend forever (blocking call).

        This method is responsible for initiating the connection to the server,
        and it must call ``bot.on_connect`` once connected, or ``bot.on_close``
        if it fails to connect.

        Upon successful connection, it must run forever, listening to the
        server and allowing the bot to use :meth:`~.send_command` in a
        thread-safe way.
        """

    def decode_line(self, line: bytes) -> str:
        """Decode a raw IRC line from ``bytes`` to ``str``."""
        # We can't trust clients to pass valid Unicode.
        try:
            data = str(line, encoding='utf-8')
        except UnicodeDecodeError as e:
            # ...unless the server announces UTF8ONLY
            if 'UTF8ONLY' in self.bot.isupport:
                raise e
            # not Unicode; let's try CP-1252
            try:
                data = str(line, encoding='cp1252')
            except UnicodeDecodeError:
                raise ValueError('Unable to decode data from server.')

        return data

    def send_command(self, *args: str, text: Optional[str] = None) -> None:
        """Send a command through the IRC connection.

        :param args: IRC command to send with its argument(s)
        :param text: the text to send (optional keyword argument)

        Example::

            # send the INFO command
            backend.send_command('INFO')
            # send the NICK command with the argument 'Sopel'
            backend.send_command('NICK', 'Sopel')
            # send the PRIVMSG command to channel #sopel with some text
            backend.send_command('PRIVMSG', '#sopel', text='Hello world!')

        .. note::

            This will call the :meth:`sopel.bot.Sopel.on_message_sent`
            callback on the bot instance with the raw message sent.
        """
        raw_command = self.prepare_command(*args, text=text)
        self.irc_send(raw_command.encode('utf-8'))
        self.bot.on_message_sent(raw_command)

    def prepare_command(self, *args: str, text: Optional[str] = None) -> str:
        """Prepare an IRC command from ``args`` and optional ``text``.

        :param args: arguments of the IRC command to send
        :param text: text to send with the IRC command (optional keyword
                     argument)
        :return: the raw message to send through the connection

        From the IRC Client Protocol specification, :rfc:`2812#section-2.3`:

            IRC messages are always lines of characters terminated with a
            CR-LF (Carriage Return - Line Feed) pair, and these messages SHALL
            NOT exceed 512 characters in length, counting all characters
            including the trailing CR-LF. Thus, there are 510 characters
            maximum allowed for the command and its parameters. There is no
            provision for continuation of message lines.

        The length in the RFC refers to the length in *bytes*, which can be
        bigger than the length of the Unicode string. This method cuts the
        message until its length fits within this limit of 510 bytes.

        The returned message contains the CR-LF pair required at the end,
        and can be sent as-is.
        """
        max_length = unicode_max_length = 510
        raw_command = ' '.join(safe(arg) for arg in args)
        if text is not None:
            raw_command = '{args} :{text}'.format(args=raw_command,
                                                  text=safe(text))

        # The max length of 512 is in bytes, not Unicode characters:
        # we can't split the message on bytes, or we may cut in the middle of a
        # multi-byte character.
        while len(raw_command.encode('utf-8')) > max_length:
            raw_command = raw_command[:unicode_max_length]
            unicode_max_length = unicode_max_length - 1

        # Ends the message with CR-LF
        return raw_command + '\r\n'

    def send_ping(self, host: str) -> None:
        """Send a ``PING`` command to the server.

        :param host: IRC server host

        A ``PING`` command should be sent at a regular interval to make sure
        the server knows the IRC connection is still active.
        """
        self.send_command('PING', host)

    def send_pong(self, host: str) -> None:
        """Send a ``PONG`` command to the server.

        :param host: IRC server host

        A ``PONG`` command must be sent each time the server sends a ``PING``
        command to the client.
        """
        self.send_command('PONG', host)

    def send_nick(self, nick: str) -> None:
        """Send a ``NICK`` command with a ``nick``.

        :param nick: nickname to take
        """
        self.send_command('NICK', nick)

    def send_user(self, user: str, mode: str, nick: str, name: str) -> None:
        """Send a ``USER`` command with a ``user``.

        :param user: IRC username
        :param mode: mode(s) to send for the user
        :param nick: nickname associated with this user
        :param name: "real name" for the user
        """
        self.send_command('USER', user, mode, nick, text=name)

    def send_pass(self, password: str) -> None:
        """Send a ``PASS`` command with a ``password``.

        :param password: password for authentication
        """
        self.send_command('PASS', password)

    def send_join(self, channel: str, password: Optional[str] = None) -> None:
        """Send a ``JOIN`` command to ``channel`` with optional ``password``.

        :param channel: channel to join
        :param password: optional password for protected channels
        """
        if password is None:
            self.send_command('JOIN', channel)
        else:
            self.send_command('JOIN', channel, password)

    def send_part(self, channel: str, reason: Optional[str] = None) -> None:
        """Send a ``PART`` command to ``channel``.

        :param channel: the channel to part
        :param text: optional text for leaving the channel
        """
        self.send_command('PART', channel, text=reason)

    def send_quit(self, reason: Optional[str] = None) -> None:
        """Send a ``QUIT`` command.

        :param reason: optional text for leaving the server

        This won't send anything if the backend isn't connected.
        """
        if self.is_connected():
            self.send_command('QUIT', text=reason)

    def send_kick(
        self,
        channel: str,
        nick: str,
        reason: Optional[str] = None,
    ) -> None:
        """Send a ``KICK`` command for ``nick`` in ``channel`` .

        :param channel: the channel from which to kick ``nick``
        :param nick: nickname to kick from the ``channel``
        :param reason: optional reason for the kick
        """
        self.send_command('KICK', channel, nick, text=reason)

    def send_privmsg(self, dest: str, text: str) -> None:
        """Send a ``PRIVMSG`` command to ``dest`` with ``text``.

        :param dest: nickname or channel name
        :param text: the text to send
        """
        self.send_command('PRIVMSG', dest, text=text)

    def send_notice(self, dest: str, text: str) -> None:
        """Send a ``NOTICE`` command to ``dest`` with ``text``.

        :param str dest: nickname or channel name
        :param str text: the text to send
        """
        self.send_command('NOTICE', dest, text=text)



"""
safety.py - Alerts about malicious URLs
Copyright © 2014, Elad Alfassa, <elad@fedoraproject.org>
Licensed under the Eiffel Forum License 2.

This plugin uses virustotal.com
"""
from __future__ import annotations

from base64 import urlsafe_b64encode
from datetime import datetime, timedelta, timezone
import json
import logging
import os.path
import re
import threading
from time import sleep
from typing import TYPE_CHECKING
from urllib.parse import urlparse, urlunparse

import requests

from sopel import plugin, tools
from sopel.config import types
from sopel.formatting import bold, color, colors

if TYPE_CHECKING:
    from typing import Optional

    from sopel.bot import Sopel, SopelWrapper
    from sopel.config import Config
    from sopel.trigger import Trigger

LOGGER = logging.getLogger(__name__)
PLUGIN_OUTPUT_PREFIX = '[safety] '

SAFETY_CACHE_KEY = "safety_cache"
SAFETY_CACHE_LOCK_KEY = SAFETY_CACHE_KEY + "_lock"
SAFETY_CACHE_LOCAL_KEY = SAFETY_CACHE_KEY + "_local"
SAFETY_MODES = ["off", "local", "local strict", "on", "strict"]
VT_API_URL = "https://www.virustotal.com/api/v3/urls"
CACHE_LIMIT = 512
known_good = []


class SafetySection(types.StaticSection):
    enabled_by_default = types.BooleanAttribute("enabled_by_default", default=True)
    """Deprecated: Sets default_mode to "off" or "on"."""
    default_mode = types.ValidatedAttribute("default_mode")
    """Which mode to use in channels without a mode set."""
    known_good = types.ListAttribute('known_good')
    """List of "known good" domains or regexes to consider trusted."""
    vt_api_key = types.ValidatedAttribute('vt_api_key')
    """Optional VirusTotal API key (improves malicious URL detection)."""
    domain_blocklist_url = types.ValidatedAttribute("domain_blocklist_url")
    """Optional hosts-file formatted domain blocklist to use instead of StevenBlack's."""


def configure(settings: Config) -> None:
    """
    | name | example | purpose |
    | ---- | ------- | ------- |
    | default\\_mode | on | Which mode to use in channels without a mode set. |
    | known\\_good | sopel.chat,dftba.net | List of "known good" domains or regexes to consider trusted. This can save VT API calls. |
    | vt\\_api\\_key | 0123456789abcdef0123456789abcdef0123456789abcdef0123456789abcdef | Optional VirusTotal API key to improve malicious URL detection |
    | domain\\_blocklist\\_url | https://example.com/bad-hosts.txt | Optional hosts-file formatted domain blocklist to use instead of StevenBlack's. |
    """
    settings.define_section("safety", SafetySection)
    settings.safety.configure_setting(
        "default_mode",
        (
            "Which mode should be used in channels that haven't specifically set one?"
            "\n({})".format("/".join(SAFETY_MODES))
        ),
        default="on",
    )
    settings.safety.configure_setting(
        'known_good',
        "Enter any domains or regexes to consider trusted",
    )
    settings.safety.configure_setting(
        'vt_api_key',
        "Optionally, enter a VirusTotal API key to improve malicious URL "
        "protection.\nOtherwise, only the configured hosts list will be used.",
    )
    settings.safety.configure_setting(
        "domain_blocklist_url",
        "Optionally, provide the URL for a hosts-file formatted domain "
        "blocklist to use instead of StevenBlack's.",
    )


def setup(bot: Sopel) -> None:
    bot.settings.define_section("safety", SafetySection)

    if bot.settings.safety.default_mode is None:
        bot.settings.safety.default_mode = "on"
    # migrate from enabled_by_default to default_mode. TODO: remove in v8.1 or v9
    if not bot.settings.safety.enabled_by_default:
        bot.settings.safety.default_mode = "off"
        LOGGER.warning(
            "config: enabled_by_default is deprecated, please use default_mode=off",
        )

    if SAFETY_CACHE_KEY not in bot.memory:
        bot.memory[SAFETY_CACHE_KEY] = tools.SopelMemory()
    if SAFETY_CACHE_LOCK_KEY not in bot.memory:
        bot.memory[SAFETY_CACHE_LOCK_KEY] = threading.Lock()
    for item in bot.settings.safety.known_good:
        known_good.append(re.compile(item, re.I))

    # clean up old files. TODO: remove in v8.1 or 9
    old_file = os.path.join(bot.settings.homedir, "malwaredomains.txt")
    if os.path.exists(old_file) and os.path.isfile(old_file):
        LOGGER.info('Removing old malwaredomains file from %s', old_file)
        try:
            os.remove(old_file)
        except Exception as err:
            # for lack of a more specific error type...
            # Python on Windows throws an exception if the file is in use
            LOGGER.warning('Could not delete %s: %s', old_file, str(err))

    LOGGER.info('Ensuring unsafe domain list is up-to-date (safety plugin setup)')
    update_local_cache(bot, init=True)


def safeify_url(url: str) -> str:
    """Replace bits of a URL to make it hard to browse to."""
    parts = urlparse(url)
    scheme = "hxx" + parts.scheme[3:]  # hxxp
    netloc = parts.netloc.replace(".", "[.]")  # google[.]com and IPv4
    netloc = netloc.replace(":", "[:]")  # IPv6 addresses (bad lazy method)
    return urlunparse((scheme, netloc) + parts[2:])


def download_domain_list(bot: Sopel, path: str) -> bool:
    """Download the current unsafe domain list.

    :param path: Where to save the unsafe domain list
    :returns: True if the list was updated
    """
    url = bot.settings.safety.domain_blocklist_url
    if url is None or not url.startswith("http"):
        url = "https://raw.githubusercontent.com/StevenBlack/hosts/master/hosts"
    LOGGER.debug("Downloading unsafe domain list from %s", url)
    old_etag = bot.db.get_plugin_value("safety", "unsafe_domain_list_etag")
    if old_etag:
        r = requests.head(url)
        if r.headers["ETag"] == old_etag and os.path.isfile(path):
            LOGGER.debug("Unsafe domain list unchanged, skipping")
            return False

    r = requests.get(url, stream=True)
    try:
        r.raise_for_status()
        with open(path + ".new", "wb") as f:
            for data in r.iter_content(None):
                f.write(data)
    except Exception:
        # don't bother handling, we'll try again tomorrow
        LOGGER.warning("Unsafe domain list download failed; using cache")
        return False
    # .new+move so we don't clobber it if the download fails in the middle
    os.rename(path + ".new", path)
    bot.db.set_plugin_value("safety", "unsafe_domain_list_etag", r.headers.get("etag"))
    return True


def update_local_cache(bot: Sopel, init: bool = False) -> None:
    """Download the current malware domain list and load it into memory.

    :param init: Load the file even if it's unchanged
    """
    path = os.path.join(bot.settings.homedir, "unsafedomains.txt")
    updated = download_domain_list(bot, path)
    if not os.path.isfile(path):
        LOGGER.warning("Could not load unsafe domain list")
        return

    if not updated and not init:
        return

    LOGGER.debug("Loading new unsafe domain list")
    unsafe_domains = set()
    with open(path, "r") as f:
        for line in f:
            clean_line = str(line).strip().lower()
            if not clean_line or clean_line[0] == '#':
                # blank line or comment
                continue

            parts = clean_line.split(' ', 1)
            try:
                domain = parts[1]
            except IndexError:
                # line does not contain a hosts entry; skip it
                continue

            if '.' in domain:
                # only publicly routable domains matter; skip loopback/link-local stuff
                unsafe_domains.add(domain)
    bot.memory[SAFETY_CACHE_LOCAL_KEY] = unsafe_domains


def shutdown(bot: Sopel) -> None:
    bot.memory.pop(SAFETY_CACHE_KEY, None)
    bot.memory.pop(SAFETY_CACHE_LOCAL_KEY, None)
    bot.memory.pop(SAFETY_CACHE_LOCK_KEY, None)


@plugin.rule(r'(?u).*(https?://\S+).*')
@plugin.priority('high')
@plugin.output_prefix(PLUGIN_OUTPUT_PREFIX)
def url_handler(bot: SopelWrapper, trigger: Trigger) -> None:
    """Checks for malicious URLs."""
    mode = bot.db.get_channel_value(
        trigger.sender,
        "safety",
        bot.settings.safety.default_mode,
    )
    if mode == "off":
        return
    local_only = "local" in mode or bot.settings.safety.vt_api_key is None
    strict = "strict" in mode

    for url in tools.web.search_urls(trigger):
        safe_url = safeify_url(url)

        positives = 0  # Number of engines saying it's malicious
        total = 0  # Number of total engines

        try:
            hostname = urlparse(url).hostname.lower()
        except (ValueError, AttributeError):
            pass  # Invalid address
        else:
            if any(regex.search(hostname) for regex in known_good):
                continue  # explicitly trusted

            if hostname in bot.memory[SAFETY_CACHE_LOCAL_KEY]:
                LOGGER.debug("[local] domain in blocklist: %r", hostname)
                positives += 1
                total += 1

        result = virustotal_lookup(bot, url, local_only=local_only)
        if result:
            positives += result["positives"]
            total += result["total"]

        if positives >= 1:
            # Possibly malicious URL detected!
            LOGGER.info(
                "Possibly malicious link (%s/%s) posted in %s by %s: %r",
                positives,
                total,
                trigger.sender,
                trigger.nick,
                safe_url,
            )
            bot.say(
                "{} {} of {} engine{} flagged a link {} posted as malicious".format(
                    bold(color("WARNING:", colors.RED)),
                    positives,
                    total,
                    "" if total == 1 else "s",
                    bold(trigger.nick),
                )
            )
            if strict:
                bot.kick(trigger.nick, trigger.sender, "Posted a malicious link")


def virustotal_lookup(
    bot: SopelWrapper,
    url: str,
    local_only: bool = False,
    max_cache_age: Optional[timedelta] = None,
) -> Optional[dict]:
    """Check VirusTotal for flags on a URL as malicious.

    :param url: The URL to look up
    :param local_only: If set, only check cache, do not make a new request
    :param max_cache_age: If set, don't use cache older than this value
    :returns: A dict containing information about findings, or None if not found
    """
    if url.startswith("hxxp"):
        url = "htt" + url[3:]
    elif not url.startswith("http"):
        # VT only does http/https URLs
        return None

    safe_url = safeify_url(url)

    # default: use any cache available
    oldest_cache = datetime(1970, 1, 1, 0, 0, tzinfo=timezone.utc)
    if max_cache_age is not None:
        oldest_cache = datetime.now(timezone.utc) - max_cache_age
    cache = bot.memory[SAFETY_CACHE_KEY]
    if url in cache and cache[url]["fetched"] > oldest_cache:
        LOGGER.debug("[VirusTotal] Using cached data for %r", safe_url)
        return bot.memory[SAFETY_CACHE_KEY].get(url)
    if local_only:
        return None

    LOGGER.debug("[VirusTotal] Looking up %r", safe_url)
    url_id = urlsafe_b64encode(url.encode("utf-8")).rstrip(b"=").decode("ascii")
    attempts = 5
    requested = False
    while attempts > 0:
        attempts -= 1
        try:
            r = requests.get(
                VT_API_URL + "/" + url_id,
                headers={"x-apikey": bot.settings.safety.vt_api_key},
            )
            if r.status_code == 200:
                vt_data = r.json()
                last_analysis = vt_data["data"]["attributes"]["last_analysis_stats"]
                # VT returns 200s for recent submissions before scan results are in...
                if not requested or sum(last_analysis.values()) > 0:
                    break
            elif not requested and r.status_code == 404:
                # Not analyzed - submit new
                LOGGER.debug("[VirusTotal] No scan for %r, requesting", safe_url)
                requests.post(
                    VT_API_URL,
                    data={"url": url},
                    headers={"x-apikey": bot.settings.safety.vt_api_key},
                )
                requested = True
                sleep(2)  # Scans seem to take ~5s minimum, so add 2s
        except requests.exceptions.RequestException:
            # Ignoring exceptions with VT so domain list will always work
            LOGGER.debug(
                "[VirusTotal] Error obtaining response for %r", safe_url, exc_info=True
            )
            return None
        except json.JSONDecodeError:
            # Ignoring exceptions with VT so domain list will always work
            LOGGER.debug(
                "[VirusTotal] Malformed response (invalid JSON) for %r",
                safe_url,
                exc_info=True,
            )
            return None
        sleep(3)
    else:  # Still no results
        LOGGER.debug("[VirusTotal] Scan failed or unfinished for %r", safe_url)
        return None
    fetched = datetime.now(timezone.utc)
    # Only count strong opinions (ignore suspicious/timeout/undetected)
    result = {
        "positives": last_analysis["malicious"],
        "total": last_analysis["malicious"] + last_analysis["harmless"],
        "fetched": fetched,
        "virustotal_data": vt_data["data"]["attributes"],
    }
    bot.memory[SAFETY_CACHE_KEY][url] = result
    if len(bot.memory[SAFETY_CACHE_KEY]) >= (2 * CACHE_LIMIT):
        _clean_cache(bot)
    return result


@plugin.command("virustotal")
@plugin.example(".virustotal https://malware.wicar.org/")
@plugin.example(".virustotal hxxps://malware.wicar.org/")
@plugin.output_prefix("[safety][VirusTotal] ")
def vt_command(bot: SopelWrapper, trigger: Trigger) -> None:
    """Look up VT results on demand."""
    if not bot.settings.safety.vt_api_key:
        bot.reply("Sorry, I don't have a VirusTotal API key configured.")
        return

    url = trigger.group(2)
    safe_url = safeify_url(url)

    result = virustotal_lookup(bot, url, max_cache_age=timedelta(minutes=1))
    if not result:
        bot.reply("Sorry, an error occurred while looking that up.")
        return

    analysis = result["virustotal_data"]["last_analysis_stats"]

    result_types = {
        "malicious": colors.RED,
        "suspicious": colors.YELLOW,
        "harmless": colors.GREEN,
        "undetected": colors.GREY,
    }
    result_strs = []
    for result_type, result_color in result_types.items():
        if analysis[result_type] == 0:
            result_strs.append("0 " + result_type)
        else:
            result_strs.append(
                bold(
                    color(str(analysis[result_type]) + " " + result_type, result_color)
                )
            )
    results_str = ", ".join(result_strs)

    vt_scan_time = datetime.fromtimestamp(
        result["virustotal_data"]["last_analysis_date"],
        timezone.utc,
    )
    bot.reply(
        "Results: {} at {} for {}".format(
            results_str,
            tools.time.format_time(
                bot.db,
                bot.config,
                nick=trigger.nick,
                channel=trigger.sender,
                time=vt_scan_time,
            ),
            safe_url,
        )
    )


@plugin.command('safety')
@plugin.example(".safety on")
@plugin.output_prefix(PLUGIN_OUTPUT_PREFIX)
def toggle_safety(bot: SopelWrapper, trigger: Trigger) -> None:
    """Set safety setting for channel."""
    if not trigger.admin and bot.channels[trigger.sender].privileges[trigger.nick] < plugin.OP:
        bot.reply('Only channel operators can change safety settings')
        return

    new_mode = None
    if trigger.group(2):
        new_mode = trigger.group(2).lower()

    if not new_mode or (new_mode != "default" and new_mode not in SAFETY_MODES):
        bot.reply(
            "Current mode: {}. Available modes: {}, or default ({})".format(
                bot.db.get_channel_value(
                    trigger.sender,
                    "safety",
                    "default",
                ),
                ", ".join(SAFETY_MODES),
                bot.settings.safety.default_mode,
            )
        )
        return

    if new_mode == "default":
        bot.db.delete_channel_value(trigger.sender, "safety")
    else:
        bot.db.set_channel_value(trigger.sender, "safety", new_mode)
    bot.say('Safety is now set to "%s" for this channel' % new_mode)


# Clean the cache every day
# Code above also calls this if there are too many cache entries
@plugin.interval(24 * 60 * 60)
def _clean_cache(bot: Sopel) -> None:
    """Cleans up old entries in URL safety cache."""

    update_local_cache(bot)

    if bot.memory[SAFETY_CACHE_LOCK_KEY].acquire(False):
        LOGGER.debug('Starting safety cache cleanup...')
        cutoff = datetime.now(timezone.utc) - timedelta(days=7)
        try:
            # clean up by age first
            old_keys = []
            for key, data in bot.memory[SAFETY_CACHE_KEY].items():
                if data['fetched'] <= cutoff:
                    old_keys.append(key)
            for key in old_keys:
                bot.memory[SAFETY_CACHE_KEY].pop(key, None)

            # clean up more values if the cache is still too big
            overage = len(bot.memory[SAFETY_CACHE_KEY]) - CACHE_LIMIT
            if overage > 0:
                extra_keys = sorted(
                    (data.fetched, key)
                    for (key, data)
                    in bot.memory[SAFETY_CACHE_KEY].items())[:overage]
                for (_, key) in extra_keys:
                    bot.memory[SAFETY_CACHE_KEY].pop(key, None)
        finally:
            # No matter what errors happen (or not), release the lock
            bot.memory[SAFETY_CACHE_LOCK_KEY].release()

        LOGGER.debug('Safety cache cleanup finished.')
    else:
        LOGGER.debug(
            'Skipping safety cache cleanup: Cache is locked, '
            'cleanup already running.')



"""
tld.py - Sopel TLD Plugin
Copyright 2009-10, Michael Yanovich, yanovich.net
Copyright 2020, dgw, technobabbl.es
Licensed under the Eiffel Forum License 2.

https://sopel.chat
"""
from __future__ import annotations

from datetime import datetime
from encodings import idna
from html.parser import HTMLParser
from json import JSONDecodeError
import logging
import re
from typing import Union

import pytz
import requests

from sopel import formatting, plugin, tools


LOGGER = logging.getLogger(__name__)


def _strptime_as_utc(time_string):
    return datetime.strptime(
        time_string, DATE_FORMAT).replace(tzinfo=pytz.utc)


DATE_FORMAT = '%Y-%m-%d %H:%M:%S'
DEFAULT_CACHE_TIME = '2000-01-01 00:00:00'
DEFAULT_CACHE_DATETIME = _strptime_as_utc(DEFAULT_CACHE_TIME)

IANA_LIST_URI = 'https://data.iana.org/TLD/tlds-alpha-by-domain.txt'
WIKI_PAGE_NAMES = [
    'List_of_Internet_top-level_domains',
    'Country_code_top-level_domain',
]
WIKI_API_PARAMS: dict[str, Union[str, int]] = {
    "action": "parse",
    "format": "json",
    "prop": "text",
    "utf8": 1,
    "formatversion": 2,
}
r_tld = re.compile(r'^\.(\S+)')
r_idn = re.compile(r'^(xn--[A-Za-z0-9]+)')


def setup(bot):
    bot.memory['tld_list_cache'] = bot.db.get_plugin_value(
        'tld', 'tld_list_cache', [])
    bot.memory['tld_list_cache_updated'] = bot.db.get_plugin_value(
        'tld', 'tld_list_cache_updated', DEFAULT_CACHE_TIME)
    bot.memory['tld_data_cache'] = bot.db.get_plugin_value(
        'tld', 'tld_data_cache', {})
    bot.memory['tld_data_cache_updated'] = bot.db.get_plugin_value(
        'tld', 'tld_data_cache_updated', DEFAULT_CACHE_TIME)

    # Restore datetime objects from string format. If parsing fails, consider
    # that cache obsolete so it will be updated at the next age check.
    try:
        bot.memory['tld_list_cache_updated'] = _strptime_as_utc(
            bot.memory['tld_list_cache_updated'])
    except ValueError:
        bot.memory['tld_list_cache_updated'] = DEFAULT_CACHE_DATETIME
    try:
        bot.memory['tld_data_cache_updated'] = _strptime_as_utc(
            bot.memory['tld_data_cache_updated'])
    except ValueError:
        bot.memory['tld_data_cache_updated'] = DEFAULT_CACHE_DATETIME


def shutdown(bot):
    if bot.memory.get('tld_list_cache'):
        bot.db.set_plugin_value(
            'tld', 'tld_list_cache', bot.memory['tld_list_cache'])
        bot.db.set_plugin_value(
            'tld', 'tld_list_cache_updated',
            bot.memory['tld_list_cache_updated'].strftime(DATE_FORMAT))
    if bot.memory.get('tld_data_cache'):
        bot.db.set_plugin_value(
            'tld', 'tld_data_cache', bot.memory['tld_data_cache'])
        bot.db.set_plugin_value(
            'tld', 'tld_data_cache_updated',
            bot.memory['tld_data_cache_updated'].strftime(DATE_FORMAT))

    for key in [
        'tld_list_cache',
        'tld_list_cache_updated',
        'tld_data_cache',
        'tld_data_cache_updated',
    ]:
        try:
            del bot.memory[key]
        except KeyError:
            pass


class WikipediaTLDListParser(HTMLParser):
    def __init__(self):
        HTMLParser.__init__(self)
        self.in_cell = False
        self.skipping = True
        self.current_row = []
        self.current_cell = ''
        self.rows = []
        self.tables = []
        self.parsed: dict[str, dict[str, str]] = {}
        self.finished = False

    def handle_starttag(self, tag, attrs):
        if tag == 'td' or tag == 'th':
            self.in_cell = True
        elif tag == 'sup' and self.in_cell:
            # ignore superscripts; they're almost exclusively footnotes
            self.skipping = True
        elif tag == 'table':
            for name, value in attrs:
                if name == 'class' and 'wikitable' in value:
                    self.skipping = False
        elif tag in ['b', 'strong'] and self.in_cell:
            self.current_cell += '<[bold]>'
        elif tag in ['i', 'em'] and self.in_cell:
            self.current_cell += '<[italic]>'

    def handle_endtag(self, tag):
        if tag == 'td' or tag == 'th':
            self.in_cell = False
            if not self.skipping:
                cell = self.current_cell.strip()
                # Python's built-in `strip()` method for strings will remove
                # some control codes we want to keep as IRC formatting. So the
                # parser inserts placeholders, and now it's time to replace
                # them with the real control codes.
                for placeholder in [
                    ('<[bold]>', formatting.CONTROL_BOLD),
                    ('<[italic]>', formatting.CONTROL_ITALIC),
                ]:
                    cell = cell.replace(*placeholder)
                self.current_row.append(cell)
            self.current_cell = ''
        elif tag == 'tr':
            if not self.skipping:
                self.rows.append(tuple(self.current_row))
            self.current_row = []
        elif tag == 'table':
            if not self.skipping:
                self.tables.append(self.rows)
            self.rows = []
            self.skipping = True
            self.in_cell = False
        elif tag == 'sup' and self.in_cell:
            self.skipping = False
        elif tag in ['b', 'strong'] and self.in_cell:
            self.current_cell += '<[bold]>'
        elif tag in ['i', 'em'] and self.in_cell:
            self.current_cell += '<[italic]>'

    def handle_data(self, data):
        if self.in_cell and not self.skipping:
            self.current_cell += data

    def get_processed_data(self):
        LOGGER.debug("Processed TLD data requested.")
        if self.finished:
            LOGGER.debug("Returning stored previously-processed data.")
            return self.parsed

        LOGGER.debug("Ensuring all buffered data has been parsed.")
        self.close()

        LOGGER.debug("Processing tables.")
        tld_list = {}
        for table in self.tables:
            headings = table[0]
            for row in table[1:]:
                key = None
                idn_key = None
                for cell in row:
                    tld = r_tld.match(cell)
                    if tld and not key:
                        key = tld.group(1).lower()
                    idn = r_idn.match(cell)
                    if idn and not idn_key:
                        idn_key = idn.group(1).lower()
                if not any([key, idn_key]):
                    LOGGER.warning(
                        "Skipping row %s; could not find string to use as lookup key.",
                        str(row),
                    )
                    continue

                # Some cleanup happens directly in the dict comprehension here.
                # Empty values (actually falsy, but only empty strings are possible)
                # and values consisting only of a dash (indicating the absence of
                # information or restrictions) get left out of the final data.
                # When the data is presented later, these empty fields are just
                # clutter taking up limited space in the IRC line.
                zipped = {
                    field: value
                    for field, value
                    in dict(zip(headings, row)).items()
                    if value and value != '—'
                }
                if key:
                    tld_list[key] = zipped
                if idn_key:
                    tld_list[idn_key] = zipped

        LOGGER.debug("Finished processing TLD data; returning it.")
        self.finished = True
        # clear working data
        del self.tables
        # cache parsed data for future requests to this parser
        self.parsed = tld_list

        return self.parsed


def _update_tld_data(bot, which, force=False):
    if which == 'list':
        then = bot.memory['tld_list_cache_updated']
    elif which == 'data':
        then = bot.memory['tld_data_cache_updated']
    else:
        LOGGER.error("Asked to update unknown cache type '%s'.", which)
        return

    now = datetime.now(tz=pytz.utc)
    since = now - then
    if not force and since.days < 7:
        LOGGER.debug(
            "Skipping TLD %s cache update; the cache is only %s old.",
            which,
            '1 day' if since.days == 1 else ('%d days' % since.days),
        )
        return

    if which == 'list':
        try:
            tld_raw = requests.get(IANA_LIST_URI).text
        except requests.exceptions.RequestException:
            # Probably a transient error; log it and continue life
            LOGGER.warning(
                "Error fetching IANA TLD list; will try again later.",
                exc_info=True)
            return

        tld_list = [
            line.lower()
            for line in tld_raw.splitlines()
            if not line.startswith('#')
        ]

        bot.memory['tld_list_cache'] = tld_list
        bot.memory['tld_list_cache_updated'] = now
    elif which == 'data':
        data_pages = []
        for title in WIKI_PAGE_NAMES:
            # don't one-liner this; dict.update() returns None, not the updated dict
            parameters = WIKI_API_PARAMS.copy()
            parameters.update({"page": title})

            try:
                # https://www.mediawiki.org/wiki/Special:MyLanguage/API:Get_the_contents_of_a_page
                tld_response = requests.get(
                    "https://en.wikipedia.org/w/api.php",
                    params=parameters,
                ).json()
                data_pages.append(tld_response["parse"]["text"])
            except (requests.exceptions.RequestException, JSONDecodeError, KeyError):
                # Log error and continue life; it'll be fine
                LOGGER.warning(
                    'Error fetching TLD data from "%s" on Wikipedia; will try again later.',
                    title, exc_info=True)
                return

        parser = WikipediaTLDListParser()
        for page in data_pages:
            parser.feed(page)

        tld_data = parser.get_processed_data()

        bot.memory['tld_data_cache'] = tld_data
        bot.memory['tld_data_cache_updated'] = now

    LOGGER.debug("Updated TLD %s cache.", which)


@plugin.interval(60 * 60)
def update_caches(bot, force=False):
    _update_tld_data(bot, 'list', force)
    _update_tld_data(bot, 'data', force)


@plugin.command('tld')
@plugin.example('.tld ru')
@plugin.output_prefix('[tld] ')
def gettld(bot, trigger):
    """Show information about the given Top Level Domain."""
    tld = trigger.group(2)
    if not tld:
        bot.reply("You must provide a top-level domain to search.")
        return  # Stop if no tld argument is provided
    tld = tld.strip('.').lower()

    if not bot.memory.get('tld_list_cache'):
        _update_tld_data(bot, 'list')
    tld_list = bot.memory['tld_list_cache']

    if not any([
        name in tld_list
        for name
        in [tld, idna.ToASCII(tld).decode('utf-8')]
    ]):
        bot.reply(
            "The top-level domain '{}' is not in IANA's list of valid TLDs."
            .format(tld))
        return

    if not bot.memory.get('tld_data_cache'):
        _update_tld_data(bot, 'data')
    tld_data = bot.memory['tld_data_cache']

    record = tld_data.get(tld, None)
    if not record:
        bot.say(
            "The top-level domain '{}' exists, "
            "but no details about it could be found."
            .format(tld)
        )
        return

    # Get the current order of available data fields
    fields = list(record.keys())
    # This trick moves matching keys to the end of the list
    fields.sort(key=lambda s: s.startswith('Notes') or s.startswith('Comments'))

    items = []
    for field in fields:
        value = record[field]
        if value:
            items.append('{}: {}'.format(field, value))

    message = ' | '.join(items)

    bot.say(message, truncation=' […]')


@plugin.command('tldcache')
@plugin.example('.tldcache clear', user_help=True)
@plugin.example('.tldcache update', user_help=True)
@plugin.example('.tldcache', user_help=True)
@plugin.require_admin
def tld_cache_command(bot, trigger):
    subcommand = trigger.group(2)
    if subcommand is None:
        times = {}
        for kind in ['list', 'data']:
            time = bot.memory.get('tld_' + kind + '_cache_updated')

            if time is None or time == DEFAULT_CACHE_DATETIME:
                times[kind] = '(not cached)'
                continue

            times[kind] = tools.time.format_time(
                db=bot.db,
                config=bot.config,
                nick=trigger.nick,
                channel=trigger.sender if not trigger.sender.is_nick() else None,
                time=time,
            )

        bot.reply(
            "IANA list updated at {}; Wikipedia data last fetched at {}."
            .format(times['list'], times['data']))
        return

    subcommand = subcommand.lower()
    if subcommand == 'update':
        bot.reply("Requesting updated IANA list and Wikipedia data.")
        update_caches(bot, force=True)
    elif subcommand == 'clear':
        bot.db.delete_plugin_value('tld', 'tld_data_cache')
        bot.memory['tld_data_cache'] = {}
        bot.db.delete_plugin_value('tld', 'tld_list_cache')
        bot.memory['tld_list_cache'] = []
        bot.db.delete_plugin_value('tld', 'tld_data_cache_updated')
        bot.memory['tld_data_cache_updated'] = DEFAULT_CACHE_DATETIME
        bot.db.delete_plugin_value('tld', 'tld_list_cache_updated')
        bot.memory['tld_list_cache_updated'] = DEFAULT_CACHE_DATETIME

        bot.reply("Cleared all cached TLD data.")
    else:
        bot.reply(
            "Unknown subcommand '{}'; recognized values are 'update' and 'clear'."
            .format(subcommand)
        )
